# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Luigi utilities
"""

from __future__ import unicode_literals, absolute_import

import warnings

from .logging import WarnSummaryAlways, WarnSummaryIfProblems

# TODO: eventually stop importing these, so that caller may import
# *some* things from this subpkg without it all breaking if luigi is
# not installed.  primary motivation here is to let the LuigiHandler
# still be created for sake of web views, where we then warn user
# that luigi is not installed.
from .tasks import (OvernightTask as OvernightTaskBase,
                    OvernightTaskWrapper as OvernightTaskWrapperBase)


class OvernightTask(OvernightTaskBase):

    def __init__(self, *args, **kwargs):
        super(OvernightTask, self).__init__(*args, **kwargs)

        warnings.warn("importing OvernightTask from `rattail.luigi` is "
                      "deprecated; please import from `rattail.luigi.tasks` "
                      "instead.", DeprecationWarning, stacklevel=2)


class OvernightTaskWrapper(OvernightTaskWrapperBase):

    def __init__(self, *args, **kwargs):
        super(OvernightTaskWrapper, self).__init__(*args, **kwargs)

        warnings.warn("importing OvernightTaskWrapper from `rattail.luigi` is "
                      "deprecated; please import from `rattail.luigi.tasks` "
                      "instead.", DeprecationWarning, stacklevel=2)



# -*- coding: utf-8 -*-
"""add product.price_required

Revision ID: 9de6ffc1d2af
Revises: 3f1e4a7b5268
Create Date: 2017-12-20 18:07:03.817694

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '9de6ffc1d2af'
down_revision = u'3f1e4a7b5268'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product
    op.add_column('product', sa.Column('price_required', sa.Boolean(), nullable=True))
    op.add_column('product_version', sa.Column('price_required', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # product
    op.drop_column('product_version', 'price_required')
    op.drop_column('product', 'price_required')



# -*- coding: utf-8 -*-
"""add purchases

Revision ID: 3350e0220faf
Revises: cd23a59c5d89
Create Date: 2016-11-05 16:12:05.364513

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3350e0220faf'
down_revision = u'cd23a59c5d89'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # purchase
    op.create_table('purchase',
                    sa.Column('store_uuid', sa.String(length=32), nullable=False),
                    sa.Column('vendor_uuid', sa.String(length=32), nullable=False),
                    sa.Column('buyer_uuid', sa.String(length=32), nullable=False),
                    sa.Column('po_number', sa.String(length=20), nullable=True),
                    sa.Column('po_total', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('date_ordered', sa.Date(), nullable=True),
                    sa.Column('date_shipped', sa.Date(), nullable=True),
                    sa.Column('date_received', sa.Date(), nullable=True),
                    sa.Column('invoice_number', sa.String(length=20), nullable=True),
                    sa.Column('invoice_date', sa.Date(), nullable=True),
                    sa.Column('invoice_total', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('status', sa.Integer(), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.ForeignKeyConstraint(['buyer_uuid'], [u'employee.uuid'], name=u'purchase_purchase_fk_buyer'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'purchase_fk_created_by'),
                    sa.ForeignKeyConstraint(['store_uuid'], [u'store.uuid'], name=u'purchase_fk_store'),
                    sa.ForeignKeyConstraint(['vendor_uuid'], [u'vendor.uuid'], name=u'purchase_fk_vendor'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # purchase_item
    op.create_table('purchase_item',
                    sa.Column('sequence', sa.Integer(), nullable=True),
                    sa.Column('vendor_code', sa.String(length=20), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=60), nullable=False),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('case_quantity', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('cases_ordered', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_ordered', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('po_unit_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('po_total', sa.Numeric(precision=7, scale=2), nullable=True),
                    sa.Column('cases_received', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_received', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('invoice_line_number', sa.Integer(), nullable=True),
                    sa.Column('invoice_case_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('invoice_unit_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('invoice_total', sa.Numeric(precision=7, scale=2), nullable=True),
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('purchase_uuid', sa.String(length=32), nullable=False),
                    sa.Column('status', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'purchase_item_fk_product'),
                    sa.ForeignKeyConstraint(['purchase_uuid'], [u'purchase.uuid'], name=u'purchase_item_fk_purchase'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # purchase_item
    op.drop_table('purchase_item')

    # purchase
    op.drop_table('purchase')



# -*- coding: utf-8; -*-
"""add batch_delproduct

Revision ID: a3a6e2f7c9a5
Revises: ef398701ffc6
Create Date: 2021-01-15 17:34:41.926477

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'a3a6e2f7c9a5'
down_revision = 'ef398701ffc6'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_delproduct
    op.create_table('batch_delproduct',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('complete', sa.Boolean(), nullable=False),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True),
                    sa.Column('extra_data', sa.Text(), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], ['user.uuid'], name='batch_delproduct_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'], name='batch_delproduct_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], ['user.uuid'], name='batch_delproduct_fk_executed_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # batch_delproduct_row
    op.create_table('batch_delproduct_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('modified', sa.DateTime(), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('item_entry', sa.String(length=32), nullable=True),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('item_id', sa.String(length=20), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('subdepartment_number', sa.Integer(), nullable=True),
                    sa.Column('subdepartment_name', sa.String(length=30), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], ['batch_delproduct.uuid'], name='batch_delproduct_row_fk_batch'),
                    sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='batch_delproduct_row_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # batch_delproduct*
    op.drop_table('batch_delproduct_row')
    op.drop_table('batch_delproduct')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendors API
"""

import warnings

from sqlalchemy.orm.exc import NoResultFound

from rattail.db import model


# TODO: remove this
def get_vendor(session, key):
    """
    Locate and return a vendor for the given key, if possible.

    .. warning::

       This function is deprecated; use
       :meth:`rattail.vendors.handler.VendorHandler.get_vendor()`
       instead.

    First the key is assumed to be a ``Vendor.id`` value.  If no matches are
    found, then it looks for a special setting in the database.  If one is
    found, ``get_vendor()`` is called again with its value.

    :param session: Active database session.

    :param key: Value to use when searching for the vendor.  

    :returns: The :class:`rattail.db.model.Vendor` instance if found; otherwise
      ``None``.
    """
    warnings.warn("api.get_vendor() is a deprecated function; please "
                  "use VendorHandler.get_vendor() method instead",
                  DeprecationWarning, stacklevel=2)

    # Vendor.uuid match?
    vendor = session.get(model.Vendor, key)
    if vendor:
        return vendor

    # Vendor.id match?
    try:
        return session.query(model.Vendor).filter_by(id=key).one()
    except NoResultFound:
        pass

    # Try settings, if value then recurse.
    from rattail.db.api import get_setting
    key = get_setting(session, 'rattail.vendor.{0}'.format(key))
    if key is None:
        return None
    return get_vendor(session, key)



# -*- coding: utf-8 -*-
"""add labels batch

Revision ID: 09976b73e96a
Revises: 9a5511fdc9f5
Create Date: 2016-11-06 16:17:14.000540

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '09976b73e96a'
down_revision = u'9a5511fdc9f5'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # label_batch
    op.create_table('label_batch',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('handheld_batch_uuid', sa.String(length=32), nullable=True),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], [u'user.uuid'], name=u'label_batch_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'label_batch_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], [u'user.uuid'], name=u'label_batch_fk_executed_by'),
                    sa.ForeignKeyConstraint(['handheld_batch_uuid'], ['batch_handheld.uuid'], name='label_batch_fk_handheld_batch'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # label_batch_row
    op.create_table('label_batch_row',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('label_code', sa.String(length=3), nullable=True),
                    sa.Column('label_profile_uuid', sa.String(length=32), nullable=True),
                    sa.Column('label_quantity', sa.Integer(), nullable=False),
                    sa.Column('location', sa.String(length=30), nullable=True),
                    sa.Column('category_code', sa.String(length=30), nullable=True),
                    sa.Column('category_name', sa.String(length=50), nullable=True),
                    sa.Column('description_2', sa.String(length=100), nullable=True),
                    sa.Column('regular_price', sa.Numeric(precision=6, scale=2), nullable=True),
                    sa.Column('pack_quantity', sa.Integer(), nullable=True),
                    sa.Column('pack_price', sa.Numeric(precision=6, scale=2), nullable=True),
                    sa.Column('sale_price', sa.Numeric(precision=6, scale=2), nullable=True),
                    sa.Column('sale_start', sa.DateTime(), nullable=True),
                    sa.Column('sale_stop', sa.DateTime(), nullable=True),
                    sa.Column('multi_unit_sale_price', sa.String(length=20), nullable=True),
                    sa.Column('comp_unit_size', sa.Numeric(precision=6, scale=2), nullable=True),
                    sa.Column('comp_unit_measure', sa.String(length=30), nullable=True),
                    sa.Column('comp_unit_reg_price', sa.String(length=10), nullable=True),
                    sa.Column('comp_unit_sale_price', sa.String(length=10), nullable=True),
                    sa.Column('vendor_id', sa.String(length=10), nullable=True),
                    sa.Column('vendor_name', sa.String(length=50), nullable=True),
                    sa.Column('vendor_item_code', sa.String(length=15), nullable=True),
                    sa.Column('case_quantity', sa.Numeric(precision=6, scale=2), nullable=True),
                    sa.Column('tax_code', sa.Integer(), nullable=True),
                    sa.Column('crv', sa.Boolean(), nullable=True),
                    sa.Column('organic', sa.Boolean(), nullable=True),
                    sa.Column('country_of_origin', sa.String(length=100), nullable=True),
                    sa.ForeignKeyConstraint(['batch_uuid'], [u'label_batch.uuid'], name=u'label_batch_row_fk_batch'),
                    sa.ForeignKeyConstraint(['label_profile_uuid'], [u'label_profile.uuid'], name=u'label_batch_row_fk_label_profile'),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'label_batch_row_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # label_batch_row
    op.drop_table('label_batch_row')

    # label_batch
    op.drop_table('label_batch')



# -*- coding: utf-8 -*-
"""add batch.complete

Revision ID: 687646b42fce
Revises: 3a500dc55e0f
Create Date: 2017-07-11 15:17:17.143593

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '687646b42fce'
down_revision = u'3a500dc55e0f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch*
    op.add_column('batch_handheld', sa.Column('complete', sa.Boolean(), nullable=True))
    op.add_column('batch_inventory', sa.Column('complete', sa.Boolean(), nullable=True))
    op.add_column('batch_pricing', sa.Column('complete', sa.Boolean(), nullable=True))
    op.add_column('label_batch', sa.Column('complete', sa.Boolean(), nullable=True))
    op.add_column('vendor_catalog', sa.Column('complete', sa.Boolean(), nullable=True))
    op.add_column('vendor_invoice', sa.Column('complete', sa.Boolean(), nullable=True))


def downgrade():

    # batch*
    op.drop_column('vendor_invoice', 'complete')
    op.drop_column('vendor_catalog', 'complete')
    op.drop_column('label_batch', 'complete')
    op.drop_column('batch_pricing', 'complete')
    op.drop_column('batch_inventory', 'complete')
    op.drop_column('batch_handheld', 'complete')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Trainwreck Handler
"""

import warnings
from collections import OrderedDict

from rattail.app import GenericHandler
from rattail.trainwreck.db import Session as TrainwreckSession


class TrainwreckHandler(GenericHandler):
    """
    Handler for Trainwreck data and databases.
    """

    def use_rotation(self):
        """
        Returns boolean indicating whether rotation should be used,
        for the Trainwreck DB(s).
        """
        return self.config.getbool('trainwreck', 'use_rotation',
                                   default=False)
    
    def current_years(self):
        """
        Returns the number of years which should be kept in the "current"
        Trainwreck DB.

        Note that this refers to the "max" number of years.  Rotation
        is done on a yearly basis, so for instance if the number of
        years to keep current is 2 (the default), then the actual
        amount of data in the current DB will vary between 1 years
        (immediately after Jan 1 pruning) and 2 years (immediately
        before Jan 1).

        In other words if today is Jan 2, and pruning has already
        occurred, then current DB would have all of last year plus
        only a day or two from this year.  But by the end of this year
        it will (still) have all of last year, plus all of this year.
        """
        return self.config.getint('trainwreck', 'current_years',
                                  default=2)

    def make_session(self, dbkey='default', **kwargs):
        """
        Make a session for a Trainwreck DB.
        """
        if 'bind' not in kwargs:
            engine = self.config.trainwreck_engines[dbkey]
            kwargs['bind'] = engine
        return TrainwreckSession(**kwargs)

    def get_model(self, **kwargs):
        """
        Return the data model for Trainwreck.
        """
        return self.config.get_trainwreck_model()

    def get_trainwreck_engines(self, include_hidden=True):
        """
        Return an "ordered" dict with configured trainwreck DB
        engines.  Keys of the dict will correspond to the config keys
        for each DB, values will be the engines.

        :param include_hidden: Flag indicating whether the result
           should include engines which are marked as hidden.  Note
           that hidden engines *are* included by default.

        :returns: An :class:`~rattail.util.OrderedDict` instance.
        """
        engines = OrderedDict(self.config.trainwreck_engines)

        if not include_hidden:
            for key in self.get_hidden_engine_keys():
                engines.pop(key, None)

        return engines

    def get_hidden_engine_keys(self):
        """
        Return a list of database engine keys which are configured to
        be hidden from the user interface.
        """
        hidden = self.config.getlist('trainwreck.db', 'hide',
                                     default=None)
        if hidden is None:
            hidden = self.config.getlist('tailbone', 'engines.trainwreck.hidden', 
                                         default=None)
            if hidden is not None:
                warnings.warn("[tailbone] 'engines.trainwreck.hidden' is a "
                              "deprecated setting, please use "
                              "[trainwreck.db] 'hide' instead",
                              DeprecationWarning)

        return hidden or []

    def engine_is_hidden(self, key):
        """
        Returns a boolean indicating if the given Trainwreck database
        engine is configured to be hidden from the user interface.
        """
        hidden = self.get_hidden_engine_keys()
        return key in hidden

    def get_oldest_transaction_date(self, session):
        """
        Query a Trainwreck database to determine the date of the
        "oldest" transaction it contains.

        :param session: SQLAlchemy session for a Trainwreck database.

        :return: A :class:`python:datetime.date` instance representing
           the oldest transaction date contained by the database.
        """
        trainwreck = self.config.get_trainwreck_model()
        txn = session.query(trainwreck.Transaction)\
                     .order_by(trainwreck.Transaction.end_time)\
                     .first()
        if txn:
            return self.app.localtime(txn.end_time, from_utc=True).date()

    def get_newest_transaction_date(self, session):
        """
        Query a Trainwreck database to determine the date of the
        "newest" transaction it contains.

        :param session: SQLAlchemy session for a Trainwreck database.

        :return: A :class:`python:datetime.date` instance representing
           the newest transaction date contained by the database.
        """
        trainwreck = self.config.get_trainwreck_model()
        txn = session.query(trainwreck.Transaction)\
                     .order_by(trainwreck.Transaction.end_time.desc())\
                     .first()
        if txn:
            return self.app.localtime(txn.end_time, from_utc=True).date()



# -*- coding: utf-8; -*-
"""add vendor terms

Revision ID: 90e3b9ea5356
Revises: dc28b97c33ff
Create Date: 2022-12-21 17:33:48.900595

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '90e3b9ea5356'
down_revision = 'dc28b97c33ff'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor
    op.add_column('vendor', sa.Column('terms', sa.String(length=100), nullable=True))
    op.add_column('vendor_version', sa.Column('terms', sa.String(length=100), autoincrement=False, nullable=True))


def downgrade():

    # vendor
    op.drop_column('vendor_version', 'terms')
    op.drop_column('vendor', 'terms')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
IFPS -> Rattail data import
"""

import datetime
import logging
from collections import OrderedDict

import sqlalchemy as sa

from rattail import importing
from rattail.importing.handlers import FromFileHandler
from rattail.importing.files import FromExcelFile
from rattail.time import localtime, make_utc
from rattail.db.util import maxlen


log = logging.getLogger(__name__)


class FromIFPSToRattail(FromFileHandler, importing.ToRattailHandler):
    """
    Handler for IFPS -> Rattail data import.
    """
    host_key = 'ifps'
    generic_host_title = "IFPS"
    host_title = "IFPS"
    local_title = "Rattail"

    def get_importers(self):
        importers = OrderedDict()
        importers['IFPS_PLU'] = IFPS_PLUImporter
        return importers


class FromIFPS(FromExcelFile):
    """
    Base class for IFPS -> Rattail data importers.
    """


class IFPS_PLUImporter(FromIFPS, importing.model.IFPS_PLUImporter):
    """
    Imports PLU data from IFPS
    """
    key = 'plu'
    supported_fields = [
        'plu',
        'category',
        'commodity',
        'variety',
        'size',
        'measurements_north_america',
        'measurements_rest_of_world',
        'restrictions_notes',
        'botanical_name',
        'aka',
        'revision_date',
        'date_added',
        'gpc',
        'image',
        'image_source',
    ]

    def normalize_host_object(self, xlrow):
        plu = int(xlrow['PLU'])
        data = {
            'plu': plu,
            'category': xlrow['CATEGORY'].strip() or None,
            'commodity': xlrow['COMMODITY'].strip() or None,
            'variety': xlrow['VARIETY'].strip() or None,
            'size': xlrow['SIZE'].strip() or None,
            'measurements_north_america': xlrow['MEASUREMENTS: NORTH AMERICA'].strip() or None,
            'measurements_rest_of_world': xlrow['MEASUREMENTS: REST OF WORLD'].strip() or None,
            'restrictions_notes': xlrow['RESTRICTIONS / NOTES'].strip() or None,
            'botanical_name': xlrow['BOTANICAL NAME'].strip() or None,
            'aka': xlrow['AKA'].strip() or None,
            'gpc': xlrow['GPC'].strip() or None,
            'image': xlrow['IMAGE'].strip() or None,
            'image_source': xlrow['IMAGE_SOURCE'].strip() or None,
        }

        # TODO: these two date fields are being interpreted as if they were in
        # the *local* time zone, but then we of course convert them to UTC when
        # importing to Rattail DB.  however, the IFPS website and data do not
        # actually indicate which time zone should apply to their values.

        if 'revision_date' in self.fields:
            dt = datetime.datetime.strptime(xlrow['REVISION DATE'], '%Y-%m-%d %H:%M:%S')
            dt = localtime(self.config, dt)
            data['revision_date'] = make_utc(dt)

        if 'date_added' in self.fields:
            dt = datetime.datetime.strptime(xlrow['DATE ADDED'], '%Y-%m-%d %H:%M:%S')
            dt = localtime(self.config, dt)
            data['date_added'] = make_utc(dt)

        # i think our schema is good enough, but just in case this it's not,
        # this should help track down any lingering issues.
        for field in self.fields:
            column = getattr(self.model_class, field)
            if isinstance(column.type, sa.String):
                length = maxlen(column)
                if length and data[field] and len(data[field]) > length:
                    log.warning("field %s is too long (%s chars) for %s: %s",
                                field, len(data[field]), plu, data[field])
                    data[field] = data[field][:length]

        return data



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tailbone API Client
"""

import json
import logging
from urllib.parse import urlparse

import requests


log = logging.getLogger(__name__)


class TailboneAPIClient(object):
    """
    Simple client for Tailbone web API.

    :param base_url: Base URL of the Tailbone API.  Usually this is
       something like ``'http://my.example.com/api'`` although YMMV.

       If you have a default URL configured as below then you do not
       need to provide a ``base_url`` to this class.

       .. code-block:: ini

          [tailbone.api]
          base_url = http://my.example.com/api

    :param max_retries: Maximum number of retries each connection
       should attempt.  This value is ultimately given to the
       :class:`~requests:requests.adapters.HTTPAdapter` instance.

       Instead of specifying this value via constructor you can add it
       to your config:

       .. code-block:: ini

          [tailbone.api]
          max_retries = 5
    """
    session = None
    logged_in = False

    def __init__(self, config, base_url=None, max_retries=None, **kwargs):
        self.config = config

        self.base_url = base_url or self.config.require(
            'tailbone.api', 'base_url')
        self.base_url = self.base_url.rstrip('/')

        if max_retries is not None:
            self.max_retries = max_retries
        else:
            self.max_retries = self.config.getint('tailbone.api', 'max_retries')

    def _init(self):
        if self.session:
            return

        self.session = requests.Session()

        # maybe *disable* SSL cert verification
        # (should only be used for testing! e.g. w/ self-signed certs)
        if not self.config.getbool('tailbone.api', 'ssl_verify', default=True):
            self.session.verify = False

        # maybe set max retries, e.g. for flaky connections
        if self.max_retries is not None:
            adapter = requests.adapters.HTTPAdapter(max_retries=self.max_retries)
            self.session.mount(self.base_url, adapter)

        # TODO: is this a good idea, or hacky security risk..?
        # without it, can get error response:
        # 400 Client Error: Bad CSRF Origin for url
        parts = urlparse(self.base_url)
        self.session.headers.update({
            'Origin': f'{parts.scheme}://{parts.netloc}',
        })

        # fetch basic 'session' endpoint, to get current xsrf token
        # (this does not require any authentication, which is next)
        response = self.get('/session')
        self.session.headers.update({
            'X-XSRF-TOKEN': response.cookies['XSRF-TOKEN']})

        # authenticate via token (preferred), or user/pass login
        token = self.config.get('tailbone.api', 'token')
        if token:
            self.session.headers.update({
                'Authorization': 'Bearer {}'.format(token),
            })
        else: # no token, so attempt login w/ credentials
            if not self.login():
                raise RuntimeError("login failed! (consider using token auth)")

    def _request(self, request_method, api_method, params=None, data=None):
        """
        Perform a request for the given API method, and return the response.
        """
        api_method = api_method.lstrip('/')
        url = '{}/{}'.format(self.base_url, api_method)
        if request_method == 'GET':
            response = self.session.get(url, params=params)
        elif request_method == 'POST':
            response = self.session.post(url, params=params,
                                         data=json.dumps(data))
        else:
            raise NotImplementedError("unknown request method: {}".format(
                request_method))
        response.raise_for_status()
        return response

    def get(self, api_method, params=None):
        """
        Perform a GET request for the given API method, and return the response.
        """
        self._init()
        return self._request('GET', api_method, params=params)

    def post(self, api_method, **kwargs):
        """
        Perform a POST request for the given API method, and return the response.
        """
        self._init()
        return self._request('POST', api_method, **kwargs)

    def login(self, username=None, password=None):
        if self.logged_in:
            return True

        if not username:
            username = self.config.require('tailbone.api', 'login.username')
        if not password:
            password = self.config.require('tailbone.api', 'login.password')

        response = self.post('/login', data={'username': username,
                                             'password': password})

        # ok means success
        data = response.json()
        if data.get('ok'):
            self.logged_in = True
            return True

        # log what we can if failure
        if data.get('error'):
            log.error("login failed: %s", data['error'])
        else:
            log.error("login failed somehow, please investigate")
        return False



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import os
import shutil
from unittest import TestCase

from rattail import util


class ImportTests(TestCase):

    def setUp(self):
        dirname = os.path.abspath(os.path.dirname(__file__))
        os.mkdir(os.path.join(dirname, 'foo'))
        with open(os.path.join(dirname, 'foo', '__init__.py'), 'w') as f:
            f.write('\n')
        with open(os.path.join(dirname, 'foo', 'bar.py'), 'w') as f:
            f.write('\n')
        os.mkdir(os.path.join(dirname, 'foo', 'baz'))
        with open(os.path.join(dirname, 'foo', 'baz', '__init__.py'), 'w') as f:
            f.write('\n')

    def tearDown(self):
        shutil.rmtree(os.path.join(os.path.dirname(__file__), 'foo'))

    def test_module_already_imported(self):
        util_module = util.import_module_path('rattail.util')
        self.assertTrue(util_module is util)

    # def test_new_module(self):
    #     dirname = os.path.abspath(os.path.dirname(__file__))

    #     foo = util.import_module_path('tests.foo')
    #     self.assertEqual(foo.__file__, os.path.abspath(
    #             os.path.join(dirname, 'foo', '__init__.py')))

    #     bar = util.import_module_path('tests.foo.bar')
    #     self.assertEqual(bar.__file__, os.path.abspath(
    #             os.path.join(dirname, 'foo', 'bar.py')))

    #     baz = util.import_module_path('tests.foo.baz')
    #     self.assertEqual(baz.__file__, os.path.abspath(
    #             os.path.join(dirname, 'foo', 'baz', '__init__.py')))

#     def test_load_object(self):
#         with open(os.path.join(os.path.dirname(__file__), 'foo', 'baz', '__init__.py'), 'w') as f:
#             f.write("""

# somevar = 42

# def somefunc():
#     return somevar * 10

# """)

#         somevar = util.load_object('tests.foo.baz:somevar')
#         self.assertEqual(somevar, 42)

#         somefunc = util.load_object('tests.foo.baz:somefunc')
#         self.assertEqual(somefunc(), 420)



# -*- coding: utf-8; -*-
"""add Person.invalid_address etc.

Revision ID: 4ff1a9a97b4e
Revises: fd91ee24cead
Create Date: 2020-01-23 15:50:25.748439

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '4ff1a9a97b4e'
down_revision = u'fd91ee24cead'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # person
    op.add_column('person', sa.Column('invalid_address', sa.Boolean(), nullable=True))
    op.add_column('person_version', sa.Column('invalid_address', sa.Boolean(), autoincrement=False, nullable=True))

    # customer
    op.add_column('customer', sa.Column('invalid_address', sa.Boolean(), nullable=True))
    op.add_column('customer_version', sa.Column('invalid_address', sa.Boolean(), autoincrement=False, nullable=True))

    # member
    op.add_column('member', sa.Column('invalid_address', sa.Boolean(), nullable=True))
    op.add_column('member_version', sa.Column('invalid_address', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # member
    op.drop_column('member_version', 'invalid_address')
    op.drop_column('member', 'invalid_address')

    # customer
    op.drop_column('customer_version', 'invalid_address')
    op.drop_column('customer', 'invalid_address')

    # person
    op.drop_column('person_version', 'invalid_address')
    op.drop_column('person', 'invalid_address')



# -*- coding: utf-8; -*-
"""migrate datasync perms

Revision ID: 24577b3bda93
Revises: 678a32b6cb19
Create Date: 2021-12-06 10:17:33.048549

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '24577b3bda93'
down_revision = '678a32b6cb19'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types


permission = sa.sql.table('permission', sa.sql.column('permission'))


def rename_perm(oldname, newname):
    op.execute(permission.update()\
               .where(permission.c.permission == oldname)\
               .values({'permission': newname}))


def upgrade():
    rename_perm('datasync.bulk_delete', 'datasync_changes.bulk_delete')
    rename_perm('datasync.delete', 'datasync_changes.delete')
    rename_perm('datasync.list', 'datasync_changes.list')
    rename_perm('datasync.view', 'datasync_changes.view')


def downgrade():
    rename_perm('datasync_changes.bulk_delete', 'datasync.bulk_delete')
    rename_perm('datasync_changes.delete', 'datasync.delete')
    rename_perm('datasync_changes.list', 'datasync.list')
    rename_perm('datasync_changes.view', 'datasync.view')



# -*- coding: utf-8; -*-
"""add dept.exempt_from_gross_sales

Revision ID: a03ee3718ff4
Revises: 05d1ed3a4a28
Create Date: 2018-06-14 11:31:36.178480

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'a03ee3718ff4'
down_revision = u'05d1ed3a4a28'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # department
    op.add_column('department', sa.Column('exempt_from_gross_sales', sa.Boolean(), nullable=True))
    op.add_column('department_version', sa.Column('exempt_from_gross_sales', sa.Boolean(), autoincrement=False, nullable=True))


def downgrade():

    # department
    op.drop_column('department_version', 'exempt_from_gross_sales')
    op.drop_column('department', 'exempt_from_gross_sales')



# -*- coding: utf-8; -*-
"""add category, family for product batch

Revision ID: 90d5e8185d44
Revises: e1685bf9f1ad
Create Date: 2019-04-23 22:41:30.384291

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '90d5e8185d44'
down_revision = 'e1685bf9f1ad'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_product_row
    op.add_column('batch_product_row', sa.Column('category_uuid', sa.String(length=32), nullable=True))
    op.add_column('batch_product_row', sa.Column('family_uuid', sa.String(length=32), nullable=True))
    op.add_column('batch_product_row', sa.Column('reportcode_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('batch_product_row_fk_category', 'batch_product_row', 'category', ['category_uuid'], ['uuid'])
    op.create_foreign_key('batch_product_row_fk_family', 'batch_product_row', 'family', ['family_uuid'], ['uuid'])
    op.create_foreign_key('batch_product_row_fk_reportcode', 'batch_product_row', 'report_code', ['reportcode_uuid'], ['uuid'])


def downgrade():

    # batch_product_row
    op.drop_constraint('batch_product_row_fk_reportcode', 'batch_product_row', type_='foreignkey')
    op.drop_constraint('batch_product_row_fk_family', 'batch_product_row', type_='foreignkey')
    op.drop_constraint('batch_product_row_fk_category', 'batch_product_row', type_='foreignkey')
    op.drop_column('batch_product_row', 'reportcode_uuid')
    op.drop_column('batch_product_row', 'family_uuid')
    op.drop_column('batch_product_row', 'category_uuid')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Work Order Handler
"""

from __future__ import unicode_literals, absolute_import

from rattail.app import GenericHandler


class WorkOrderHandler(GenericHandler):
    """
    Handler for work orders.
    """

    def make_workorder(self, session, **kwargs):
        """
        Make and return a new work order.
        """
        model = self.model

        if 'id' not in kwargs:
            kwargs['id'] = self.app.next_counter_value(session, 'workorder_id')

        if 'date_submitted' not in kwargs:
            kwargs['date_submitted'] = self.app.today()

        if 'status_code' not in kwargs:
            kwargs['status_code'] = self.enum.WORKORDER_STATUS_SUBMITTED

        workorder = model.WorkOrder(**kwargs)
        session.add(workorder)
        self.record_event(workorder, self.enum.WORKORDER_EVENT_SUBMITTED)
        session.flush()
        return workorder

    def status_codes(self):
        """
        Retrieve all info about possible work order status codes.
        """
        code_names = {}
        for name in dir(self.enum):
            if name.startswith('WORKORDER_STATUS_'):
                code_names[getattr(self.enum, name)] = name

        status_codes = []
        for key, label in self.enum.WORKORDER_STATUS.items():
            status_codes.append({
                'code': key,
                'code_name': code_names[key],
                'label': label,
            })

        return status_codes

    def record_event(self, workorder, type_code, **kwargs):
        model = self.model

        # who did this?
        if 'user' in kwargs:
            user = kwargs['user']
        else:
            session = kwargs.get('session') or self.app.get_session(workorder)
            user = session.continuum_user
        kwargs['user'] = user

        # record the event
        workorder.events.append(model.WorkOrderEvent(
            type_code=type_code, **kwargs))

    def receive(self, workorder, **kwargs):
        """
        Sets work order status to "received".
        """
        workorder.status_code = self.enum.WORKORDER_STATUS_RECEIVED
        workorder.date_received = self.app.today()
        self.record_event(workorder, self.enum.WORKORDER_EVENT_RECEIVED)

    def await_estimate(self, workorder):
        """
        Sets work order status to "awaiting estimate confirmation".
        """
        workorder.status_code = self.enum.WORKORDER_STATUS_PENDING_ESTIMATE
        self.record_event(workorder, self.enum.WORKORDER_EVENT_PENDING_ESTIMATE)

    def await_parts(self, workorder):
        """
        Sets work order status to "awaiting parts".
        """
        workorder.status_code = self.enum.WORKORDER_STATUS_WAITING_FOR_PARTS
        self.record_event(workorder, self.enum.WORKORDER_EVENT_WAITING_FOR_PARTS)

    def work_on_it(self, workorder):
        """
        Sets work order status to "working on it".
        """
        workorder.status_code = self.enum.WORKORDER_STATUS_WORKING_ON_IT
        self.record_event(workorder, self.enum.WORKORDER_EVENT_WORKING_ON_IT)

    def release(self, workorder):
        """
        Sets work order status to "released".
        """
        workorder.status_code = self.enum.WORKORDER_STATUS_RELEASED
        workorder.date_released = self.app.today()
        self.record_event(workorder, self.enum.WORKORDER_EVENT_RELEASED)

    def deliver(self, workorder):
        """
        Sets work order status to "delivered".
        """
        workorder.status_code = self.enum.WORKORDER_STATUS_DELIVERED
        workorder.date_delivered = self.app.today()
        self.record_event(workorder, self.enum.WORKORDER_EVENT_DELIVERED)

    def cancel(self, workorder):
        """
        Sets work order status to "canceled".
        """
        workorder.status_code = self.enum.WORKORDER_STATUS_CANCELED
        self.record_event(workorder, self.enum.WORKORDER_EVENT_CANCELED)



# -*- coding: utf-8; -*-
"""rename batch_custorder

Revision ID: 6b06ebea979c
Revises: 6a691bede834
Create Date: 2021-02-02 14:51:10.864672

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '6b06ebea979c'
down_revision = '6a691bede834'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # remove references to `custorder_batch`
    op.drop_constraint('custorder_batch_row_fk_batch_uuid', 'custorder_batch_row', type_='foreignkey')

    # batch_custorder
    op.drop_constraint('custorder_batch_fk_created_by', 'custorder_batch', type_='foreignkey')
    op.drop_constraint('custorder_batch_fk_cognized_by', 'custorder_batch', type_='foreignkey')
    op.drop_constraint('custorder_batch_fk_executed_by', 'custorder_batch', type_='foreignkey')
    op.drop_constraint('custorder_batch_fk_store', 'custorder_batch', type_='foreignkey')
    op.drop_constraint('custorder_batch_fk_customer', 'custorder_batch', type_='foreignkey')
    op.drop_constraint('custorder_batch_fk_person', 'custorder_batch', type_='foreignkey')
    op.drop_constraint('custorder_batch_fk_order', 'custorder_batch', type_='foreignkey')
    op.rename_table('custorder_batch', 'batch_custorder')
    op.create_foreign_key('batch_custorder_fk_created_by',
                          'batch_custorder', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_custorder_fk_cognized_by',
                          'batch_custorder', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_custorder_fk_executed_by',
                          'batch_custorder', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('batch_custorder_fk_store',
                          'batch_custorder', 'store',
                          ['store_uuid'], ['uuid'])
    op.create_foreign_key('batch_custorder_fk_customer',
                          'batch_custorder', 'customer',
                          ['customer_uuid'], ['uuid'])
    op.create_foreign_key('batch_custorder_fk_person',
                          'batch_custorder', 'person',
                          ['person_uuid'], ['uuid'])
    op.create_foreign_key('batch_custorder_fk_order',
                          'batch_custorder', 'custorder',
                          ['order_uuid'], ['uuid'])

    # batch_custorder_row
    op.drop_constraint('custorder_batch_row_fk_product', 'custorder_batch_row', type_='foreignkey')
    op.drop_constraint('custorder_batch_row_fk_item', 'custorder_batch_row', type_='foreignkey')
    op.rename_table('custorder_batch_row', 'batch_custorder_row')
    op.create_foreign_key('batch_custorder_row_fk_batch_uuid',
                          'batch_custorder_row', 'batch_custorder',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('batch_custorder_row_fk_product',
                          'batch_custorder_row', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('batch_custorder_row_fk_item',
                          'batch_custorder_row', 'custorder_item',
                          ['item_uuid'], ['uuid'])

def downgrade():

    # remove references to `batch_custorder`
    op.drop_constraint('batch_custorder_row_fk_batch_uuid', 'batch_custorder_row', type_='foreignkey')

    # batch_custorder
    op.drop_constraint('batch_custorder_fk_created_by', 'batch_custorder', type_='foreignkey')
    op.drop_constraint('batch_custorder_fk_cognized_by', 'batch_custorder', type_='foreignkey')
    op.drop_constraint('batch_custorder_fk_executed_by', 'batch_custorder', type_='foreignkey')
    op.drop_constraint('batch_custorder_fk_store', 'batch_custorder', type_='foreignkey')
    op.drop_constraint('batch_custorder_fk_customer', 'batch_custorder', type_='foreignkey')
    op.drop_constraint('batch_custorder_fk_person', 'batch_custorder', type_='foreignkey')
    op.drop_constraint('batch_custorder_fk_order', 'batch_custorder', type_='foreignkey')
    op.rename_table('batch_custorder', 'custorder_batch')
    op.create_foreign_key('custorder_batch_fk_created_by',
                          'custorder_batch', 'user',
                          ['created_by_uuid'], ['uuid'])
    op.create_foreign_key('custorder_batch_fk_cognized_by',
                          'custorder_batch', 'user',
                          ['cognized_by_uuid'], ['uuid'])
    op.create_foreign_key('custorder_batch_fk_executed_by',
                          'custorder_batch', 'user',
                          ['executed_by_uuid'], ['uuid'])
    op.create_foreign_key('custorder_batch_fk_store',
                          'custorder_batch', 'store',
                          ['store_uuid'], ['uuid'])
    op.create_foreign_key('custorder_batch_fk_customer',
                          'custorder_batch', 'customer',
                          ['customer_uuid'], ['uuid'])
    op.create_foreign_key('custorder_batch_fk_person',
                          'custorder_batch', 'person',
                          ['person_uuid'], ['uuid'])
    op.create_foreign_key('custorder_batch_fk_order',
                          'custorder_batch', 'custorder',
                          ['order_uuid'], ['uuid'])

    # batch_custorder_row
    op.drop_constraint('batch_custorder_row_fk_product', 'batch_custorder_row', type_='foreignkey')
    op.drop_constraint('batch_custorder_row_fk_item', 'batch_custorder_row', type_='foreignkey')
    op.rename_table('batch_custorder_row', 'custorder_batch_row')
    op.create_foreign_key('custorder_batch_row_fk_batch_uuid',
                          'custorder_batch_row', 'custorder_batch',
                          ['batch_uuid'], ['uuid'])
    op.create_foreign_key('custorder_batch_row_fk_product',
                          'custorder_batch_row', 'product',
                          ['product_uuid'], ['uuid'])
    op.create_foreign_key('custorder_batch_row_fk_item',
                          'custorder_batch_row', 'custorder_item',
                          ['item_uuid'], ['uuid'])



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth Handler

See also :doc:`rattail-manual:base/handlers/other/auth`.
"""

import secrets
import warnings

from sqlalchemy import orm
import sqlalchemy_continuum as continuum

from rattail.app import GenericHandler, MergeMixin


class AuthHandler(GenericHandler, MergeMixin):
    """
    Base class and default implementation for the so-called "auth"
    handler, by which we mean "authentication and authorization".

    In practice this is also responsible for creating new users, and
    various things pertaining to roles, etc.
    """

    def authenticate_user(self, session, username, password):
        """
        Authenticate the given user credentials, and if successful,
        return the user object.

        Default logic will (try to) locate a user record with matching
        username, then confirm the supplied password is also a match.

        You may of course define a custom handler and then could
        authenticate against anything you like, e.g. the POS system or
        LDAP etc.  The only trick is that this must return a Rattail
        user, not some other kind.  So you may have to devisde a way
        to auto-create the Rattail user as needed, when authentication
        for the external system succeeds.

        Generally speaking the credentials passed in will have come
        directly from a user login attempt in the web app etc.  Again
        the default logic assumes a "username" but in practice it may
        be an email address etc. - whatever the user types.

        :param session: Current session for Rattail DB.

        :param username: Username as string.

        :param password: Password as string.

        :returns: On success, a :class:`~rattail.db.model.users.User`
           instance; else ``None``.
        """
        from rattail.db.auth import authenticate_user

        return authenticate_user(session, username, password)

    def authenticate_user_token(self, session, token):
        """
        Authenticate the given user API token string, and if valid,
        return the corresponding User object.
        """
        model = self.model

        try:
            token = session.query(model.UserAPIToken)\
                           .filter(model.UserAPIToken.token_string == token)\
                           .one()
        except orm.exc.NoResultFound:
            pass
        else:
            user = token.user
            if user.active:
                return user

    def has_permission(self, session, principal, permission,
                       include_guest=True,
                       include_authenticated=True):
        """
        Check if the given user or role has been granted the given
        permission.

        :param session: Current session for Rattail DB.

        :param principal: Either a
           :class:`~rattail.db.model.users.User` or
           :class:`~rattail.db.model.users.Role` instance.  It is also
           expected that this may sometimes be ``None``, in which case
           the "Guest" role will typically be assumed.

        :param permission: Name of the permission for which to check.

        :param include_guest: Whether or not the "Guest" role should
           be included when checking permissions.  If ``False``, then
           Guest's permissions will *not* be consulted.

        :param include_authenticated: Whether or not the
           "Authenticated" role should be included when checking
           permissions.

        :returns: Boolean indicating if the permission has been
           granted.
        """
        perms = self.get_permissions(session, principal,
                                     include_guest=include_guest,
                                     include_authenticated=include_authenticated)
        return permission in perms

    def get_permissions(self, session, principal,
                          include_guest=True,
                          include_authenticated=True):
        """
        Return a set of permission names, which represents all
        permissions effectively granted to the given user or role.

        :param session: Current session for Rattail DB.

        :param principal: Either a
           :class:`~rattail.db.model.users.User` or
           :class:`~rattail.db.model.users.Role` instance.  It is also
           expected that this may sometimes be ``None``, in which case
           the "Guest" role will typically be assumed.

        :param include_guest: Whether or not the "Guest" role should
           be included when checking permissions.  If ``False``, then
           Guest's permissions will *not* be consulted.

        :param include_authenticated: Whether or not the
           "Authenticated" role should be included when checking
           permissions.

        :returns: Set of permission names.
        """
        from rattail.db.auth import guest_role, authenticated_role

        # we will use any `roles` attribute which may be present.  in practice we
        # would be assuming a User in this case
        if hasattr(principal, 'roles'):

            roles = []
            for role in principal.roles:
                include = False
                if role.node_type:
                    if role.node_type == self.config.node_type():
                        include = True
                else:
                    include = True
                if include:
                    roles.append(role)

            # here our User assumption gets a little more explicit
            if include_authenticated:
                roles.append(authenticated_role(session))

        # otherwise a non-null principal is assumed to be a Role
        elif principal is not None:
            roles = [principal]

        # fallback assumption is "no roles"
        else:
            roles = []

        # maybe include guest roles
        if include_guest:
            roles.append(guest_role(session))

        # build the permissions cache
        cache = set()
        for role in roles:
            cache.update(role.permissions)

        return cache

    def cache_permissions(self, *args, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "get_permissions() method instead",
                      DeprecationWarning)
        return self.get_permissions(*args, **kwargs)

    def grant_permission(self, role, permission):
        """
        Grant a permission to the role.  If the role already has the
        permission, nothing is done.

        :param role: A :class:`~rattail.db.model.users.Role` instance.

        :param permission: Name of the permission as string.
        """
        if permission not in role.permissions:
            role.permissions.append(permission)

    def revoke_permission(self, role, permission):
        """
        Revoke a permission from the role.  If the role does not have
        the permission, nothing is done.

        :param role: A :class:`~rattail.db.model.users.Role` instance.

        :param permission: Name of the permission as string.
        """
        if permission in role.permissions:
            role.permissions.remove(permission)

    def generate_preferred_username(self, session, **kwargs):
        """
        Generate a "preferred" username using data from ``kwargs`` as
        hints.

        Note that ``kwargs`` should be of the same sort that might be
        passed to the constructor for a new
        :class:`~rattail.db.model.users.User` instance.

        So far there is only one "hint" which is honored by the
        default logic; however the intention is to leave this flexible
        as other kinds of hints may be useful in the future.

        This method does not confirm if the username it generates is
        actually "available" for a new user.  If you need confirmation
        then use :meth:`generate_unique_username()` instead.

        :param session: Current session for Rattail DB.

        :param person: Reference to a
           :class:`~rattail.db.model.people.Person` instance.  If you
           specify this hint, then default logic will generate a
           username using first and last names, like ``'first.last'``.
           (You can override with a custom handler if needed.)

        :returns: Generated username as string.
        """
        person = kwargs.get('person')
        if person:
            first = (person.first_name or '').strip().lower()
            last = (person.last_name or '').strip().lower()
            return '{}.{}'.format(first, last)

        return 'newuser'

    def generate_username(self, *args, **kwargs): # pragma: no cover
        warnings.warn("method is deprecated, please use "
                      "generate_preferred_username() method instead",
                      DeprecationWarning)
        return self.generate_preferred_username(*args, **kwargs)

    def generate_unique_username(self, session, **kwargs):
        """
        Generate a *unique* username using data from ``kwargs`` as
        hints.

        Note that ``kwargs`` should be of the same sort that might be
        passed to the constructor for a new
        :class:`~rattail.db.model.users.User` instance.

        This method is a convenience which does two things:

        First it calls :meth:`generate_preferred_username()` to obtain
        the "preferred" username.  (It passes ``kwargs`` along when it
        makes the call.  See :meth:`generate_preferred_username()` for
        more info.)

        Then it checks to see if the resulting username is already
        taken.  If it is, then a "counter" is appended to the
        username, and incremented until a username can be found which
        is *not* yet taken.

        It returns the first "available" (hence unique) username which
        is found.  Note that it is considered unique and therefore
        available *at the time*; however this method does not
        "reserve" the username in any way.  It is assumed that you
        would create the user yourself once you have the username.

        :param session: Current session for Rattail DB.

        :returns: Username as string.
        """
        model = self.model

        original_username = self.generate_preferred_username(session, **kwargs)
        username = original_username

        # only if given a session, can we check for unique username
        if session:
            counter = 1
            while True:
                users = session.query(model.User)\
                               .filter(model.User.username == username)\
                               .count()
                if not users:
                    break
                username = "{}{:02d}".format(original_username, counter)
                counter += 1

        return username

    def make_user(self, session=None, **kwargs):
        """
        Make and return a new user.

        This is mostly just a simple wrapper around the normal
        :class:`~rattail.db.model.users.User` constructor.  All
        ``kwargs`` for instance are passed on to the constructor.

        Default logic here only adds one other convenience:

        If there is no ``username`` specified in the ``kwargs`` then
        it will call :meth:`generate_unique_username()` to
        automatically provide a username.  Note that all ``kwargs``
        are passed along in that call.

        :param session: Current session for the Rattail DB.  This is
           "sort of" optional, but please do provide it, as it may
           become requied in the future.

        :returns: A new :class:`~rattail.db.model.users.User` instance.
        """
        model = self.model

        if 'username' not in kwargs:
            kwargs['username'] = self.generate_unique_username(session, **kwargs)

        user = model.User(**kwargs)
        if session:
            session.add(user)
        return user

    def get_email_address(self, user, **kwargs):
        """
        Get the "best" email address we have on file for the given user.
        """
        if user.person:
            for email in user.person.emails:
                if not email.invalid:
                    return email.address

        # TODO: remove this, legacy code
        return user.get_email_address()

    def generate_raw_api_token(self):
        """
        Generate a new *raw* API token string.
        """
        return secrets.token_urlsafe()

    def add_api_token(self, user, description, **kwargs):
        """
        Add a new API token for the user.
        """
        model = self.model
        session = self.app.get_session(user)

        # generate raw API token, in the form required for use within
        # the API client
        token_string = self.generate_raw_api_token()

        # create DB record for the token
        token = model.UserAPIToken(
            user=user,
            description=description,
            token_string=token_string)
        session.add(token)

        return token

    def delete_api_token(self, token, **kwargs):
        """
        Delete a new API token for the user.
        """
        session = self.app.get_session(token)
        session.delete(token)

    def get_merge_preview_fields(self, **kwargs):
        """
        Returns a sequence of fields which will be used during a merge
        preview.
        """
        F = self.make_merge_field
        return [
            F('uuid'),
            F('username'),
            F('person_uuid', coalesce=True),
            F('person_name', coalesce=True),
            F('role_count'),    # coalesced manually
            F('active', coalesce=True),
            F('sent_message_count', additive=True),
            F('received_message_count', additive=True),
        ]

    def get_merge_preview_data(self, user, **kwargs):
        return {
            'uuid': user.uuid,
            'username': user.username,
            'person_uuid': user.person_uuid,
            'person_name': user.person.display_name if user.person else None,
            '_roles': user.roles, # needed for final role count
            'role_count': len(user.roles),
            'active': user.active,
            'sent_message_count': len(user.sent_messages),
            'received_message_count': len(user._messages),
        }

    def get_merge_resulting_data(self, removing, keeping, **kwargs):
        result = super(AuthHandler, self).get_merge_resulting_data(
            removing, keeping, **kwargs)

        # nb. must "manually" coalesce the role count
        result['role_count'] = len(set(removing['_roles'] + keeping['_roles']))

        return result

    def why_not_merge(self, removing, keeping, **kwargs):

        if removing.sent_messages:
            return "Cannot (yet) remove a user who has sent messages"

        if removing._messages:
            return "Cannot (yet) remove a user who has received messages"

        if removing._roles:
            return "Cannot (yet) remove a user who is assigned to roles"

    def merge_update_keeping_object(self, removing, keeping):
        super(AuthHandler, self).merge_update_keeping_object(removing, keeping)
        session = self.app.get_session(keeping)
        model = self.model

        # update any notes authored by old user, to reflect new user
        notes = session.query(model.Note)\
                       .filter(model.Note.created_by == removing)\
                       .all()
        for note in notes:
            note.created_by = keeping

    def delete_user(self, user, **kwargs):
        """
        Delete the given user account.  Use with caution!  As this
        generally cannot be undone.

        Default behavior here is of course to delete the account, but
        it also must try to "remove" the user association from various
        places, in particular the continuum transactions table.
        Please note that this will leave certain record versions as
        appearing to be "without an author".

        :param user: Reference to a
           :class:`~rattail.db.model.users.User` to be deleted.

        :returns: Boolean indicating success.

           Note that the utility of this method even having a return
           value is deemed questionable, so it's possible in the
           future this may just return ``None`` on success, and raise
           an error to indicate failure.
        """
        session = self.app.get_session(user)

        # disassociate user from transactions
        if self.config.versioning_has_been_enabled:
            self.remove_user_from_continuum_transactions(user)

        # finally, delete the user outright
        session.delete(user)
        return True

    def remove_user_from_continuum_transactions(self, user):
        """
        Remove the given user from all Continuum transactions,
        i.e. all data versioning tables.

        You probably will not need to invoke this directly; it is
        invoked as needed from within :meth:`delete_user()`.

        :param user: A :class:`~rattail.db.model.users.User` instance
           which should be purged from the versioning tables.
        """
        session = self.app.get_session(user)
        model = self.model

        # remove the user from any continuum transactions
        # nb. we can use "any" model class here, to obtain Transaction
        Transaction = continuum.transaction_class(model.User)
        transactions = session.query(Transaction)\
                              .filter(Transaction.user_id == user.uuid)\
                              .all()
        for txn in transactions:
            txn.user_id = None



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Generator for 'tailbone-integration' projects
"""

import os

import colander

from rattail.projects import PythonProjectGenerator
from rattail.util import get_studly_prefix, get_package_name


class TailboneIntegrationProjectGenerator(PythonProjectGenerator):
    """
    Generator for projects which integrate Tailbone with some other
    system.  This is for generating projects such as tailbone-corepos
    and tailbone-mailchimp etc.
    """
    key = 'tailbone_integration'

    def make_schema(self, **kwargs):
        schema = super(TailboneIntegrationProjectGenerator, self).make_schema(**kwargs)

        schema.add(colander.SchemaNode(name='integration_name',
                                       typ=colander.String()))

        schema.add(colander.SchemaNode(name='integration_url',
                                       typ=colander.String()))

        schema.add(colander.SchemaNode(name='has_static_files',
                                       typ=colander.Boolean()))

        return schema

    def normalize_context(self, context):
        context = super(TailboneIntegrationProjectGenerator, self).normalize_context(context)

        if not context.get('description'):
            context['description'] = "Tailbone integration package for {}".format(
                context['integration_name'])

        context['classifiers'].update(set([
            'Environment :: Web Environment',
            'Framework :: Pyramid',
            'Operating System :: POSIX :: Linux',
            'Topic :: Office/Business',
        ]))

        context['requires']['Tailbone'] = True

        if 'integration_studly_prefix' not in context:
            context['integration_studly_prefix'] = get_studly_prefix(
                context['integration_name'])

        if 'integration_pkgname' not in context:
            context['integration_pkgname'] = get_package_name(
                context['integration_name'])

        if 'year' not in context:
            context['year'] = self.app.today().year

        return context

    def generate_project(self, output, context, **kwargs):
        super(TailboneIntegrationProjectGenerator, self).generate_project(
            output, context, **kwargs)

        package = os.path.join(output, context['pkg_name'])

        ##############################
        # views
        ##############################

        views = os.path.join(package, 'views')
        os.makedirs(views)

        self.generate('package/views/__init__.py',
                      os.path.join(views, '__init__.py'))

        ##############################
        # static
        ##############################

        if context['has_static_files']:

            static = os.path.join(package, 'static')
            os.makedirs(static)

            self.generate('package/static/__init__.py.mako',
                          os.path.join(static, '__init__.py'),
                          context)

        ##############################
        # templates
        ##############################

        templates = os.path.join(package, 'templates')
        os.makedirs(templates)

        self.generate('package/templates/.keepme',
                      os.path.join(templates, '.keepme'))



# -*- coding: utf-8 -*-
"""add purchase batch

Revision ID: 9a5511fdc9f5
Revises: 3350e0220faf
Create Date: 2016-11-05 16:33:00.953452

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '9a5511fdc9f5'
down_revision = u'3350e0220faf'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # purchase_batch
    op.create_table('purchase_batch',
                    sa.Column('store_uuid', sa.String(length=32), nullable=False),
                    sa.Column('vendor_uuid', sa.String(length=32), nullable=False),
                    sa.Column('buyer_uuid', sa.String(length=32), nullable=False),
                    sa.Column('po_number', sa.String(length=20), nullable=True),
                    sa.Column('po_total', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('date_ordered', sa.Date(), nullable=True),
                    sa.Column('date_shipped', sa.Date(), nullable=True),
                    sa.Column('date_received', sa.Date(), nullable=True),
                    sa.Column('invoice_number', sa.String(length=20), nullable=True),
                    sa.Column('invoice_date', sa.Date(), nullable=True),
                    sa.Column('invoice_total', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.ForeignKeyConstraint(['buyer_uuid'], [u'employee.uuid'], name=u'purchase_batch_purchase_fk_buyer'),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], [u'user.uuid'], name=u'purchase_batch_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'purchase_batch_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], [u'user.uuid'], name=u'purchase_batch_fk_executed_by'),
                    sa.ForeignKeyConstraint(['store_uuid'], [u'store.uuid'], name=u'purchase_batch_fk_store'),
                    sa.ForeignKeyConstraint(['vendor_uuid'], [u'vendor.uuid'], name=u'purchase_batch_fk_vendor'),
                    sa.PrimaryKeyConstraint('uuid')
    )

    # purchase_batch_row
    op.create_table('purchase_batch_row',
                    sa.Column('vendor_code', sa.String(length=20), nullable=True),
                    sa.Column('product_uuid', sa.String(length=32), nullable=True),
                    sa.Column('upc', rattail.db.types.GPCType(), nullable=True),
                    sa.Column('brand_name', sa.String(length=100), nullable=True),
                    sa.Column('description', sa.String(length=60), nullable=False),
                    sa.Column('size', sa.String(length=255), nullable=True),
                    sa.Column('department_number', sa.Integer(), nullable=True),
                    sa.Column('department_name', sa.String(length=30), nullable=True),
                    sa.Column('case_quantity', sa.Numeric(precision=8, scale=2), nullable=True),
                    sa.Column('cases_ordered', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_ordered', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('po_unit_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('po_total', sa.Numeric(precision=7, scale=2), nullable=True),
                    sa.Column('cases_received', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('units_received', sa.Numeric(precision=10, scale=4), nullable=True),
                    sa.Column('invoice_line_number', sa.Integer(), nullable=True),
                    sa.Column('invoice_case_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('invoice_unit_cost', sa.Numeric(precision=7, scale=3), nullable=True),
                    sa.Column('invoice_total', sa.Numeric(precision=7, scale=2), nullable=True),
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('batch_uuid', sa.String(length=32), nullable=False),
                    sa.Column('sequence', sa.Integer(), nullable=False),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('removed', sa.Boolean(), nullable=False),
                    sa.ForeignKeyConstraint(['batch_uuid'], [u'purchase_batch.uuid'], name=u'purchase_batch_row_fk_batch_uuid'),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'purchase_batch_row_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    

def downgrade():

    # purchase_batch_row
    op.drop_table('purchase_batch_row')

    # purchase_batch
    op.drop_table('purchase_batch')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2023 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Generator for 'byjove' projects
"""

import os

import colander

from rattail.projects import ProjectGenerator


class ByjoveProjectGenerator(ProjectGenerator):
    """
    Generator for Byjove app projects.
    """
    key = 'byjove'

    def make_schema(self, **kwargs):
        schema = colander.Schema()

        schema.add(colander.SchemaNode(name='system_name',
                                       typ=colander.String()))

        schema.add(colander.SchemaNode(name='name',
                                       typ=colander.String()))

        schema.add(colander.SchemaNode(name='slug',
                                       typ=colander.String()))

        return schema

    def generate_project(self, output, context, **kwargs):

        ##############################
        # root project dir
        ##############################

        self.generate('CHANGELOG.md.mako',
                      os.path.join(output, 'CHANGELOG.md'),
                      context)

        self.generate('gitignore',
                      os.path.join(output, '.gitignore'))

        self.generate('README.md.mako',
                      os.path.join(output, 'README.md'),
                      context)

        self.generate('vue.config.js.dist.mako',
                      os.path.join(output, 'vue.config.js.dist'),
                      context)

        self.generate('package.json.mako',
                      os.path.join(output, 'package.json'),
                      context)

        self.generate('tasks.py.mako',
                      os.path.join(output, 'tasks.py'),
                      context)

        ##############################
        # public
        ##############################

        public = os.path.join(output, 'public')
        os.makedirs(public)

        self.generate('public/index.html.mako',
                      os.path.join(public, 'index.html'),
                      context)

        self.generate('public/favicon.ico',
                      os.path.join(public, 'favicon.ico'))

        ##############################
        # src
        ##############################

        src = os.path.join(output, 'src')
        os.makedirs(src)

        self.generate('src/appsettings.js.dist.mako',
                      os.path.join(src, 'appsettings.js.dist'),
                      context)

        self.generate('src/appsettings.production.js.mako',
                      os.path.join(src, 'appsettings.production.js'),
                      context)

        self.generate('src/App.vue',
                      os.path.join(src, 'App.vue'))

        self.generate('src/main.js',
                      os.path.join(src, 'main.js'))

        self.generate('src/router.js',
                      os.path.join(src, 'router.js'))

        self.generate('src/store.js',
                      os.path.join(src, 'store.js'))

        ##############################
        # assets
        ##############################

        assets = os.path.join(src, 'assets')
        os.makedirs(assets)

        self.generate('src/assets/Hymenocephalus_italicus.jpg',
                      os.path.join(assets, 'Hymenocephalus_italicus.jpg'))

        ##############################
        # components
        ##############################

        components = os.path.join(src, 'components')
        os.makedirs(components)

        self.generate('src/components/AppNav.vue',
                      os.path.join(components, 'AppNav.vue'))

        ##############################
        # views
        ##############################

        views = os.path.join(src, 'views')
        os.makedirs(views)

        self.generate('src/views/Home.vue',
                      os.path.join(views, 'Home.vue'))

        self.generate('src/views/Login.vue',
                      os.path.join(views, 'Login.vue'))

        self.generate('src/views/About.vue',
                      os.path.join(views, 'About.vue'))

        ###############
        # customers
        ###############

        customers = os.path.join(views, 'customers')
        os.makedirs(customers)

        self.generate('src/views/customers/Customers.vue',
                      os.path.join(customers, 'Customers.vue'))

        self.generate('src/views/customers/Customer.vue',
                      os.path.join(customers, 'Customer.vue'))

        self.generate('src/views/customers/index.js',
                      os.path.join(customers, 'index.js'))

        ###############
        # products
        ###############

        products = os.path.join(views, 'products')
        os.makedirs(products)

        self.generate('src/views/products/Products.vue',
                      os.path.join(products, 'Products.vue'))

        self.generate('src/views/products/Product.vue',
                      os.path.join(products, 'Product.vue'))

        self.generate('src/views/products/index.js',
                      os.path.join(products, 'index.js'))



# -*- coding: utf-8; -*-
"""add batch_delproduct.inactiviy_months

Revision ID: e54234dcce5f
Revises: 0e37a0b5dc9a
Create Date: 2021-03-09 11:35:30.446866

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'e54234dcce5f'
down_revision = '0e37a0b5dc9a'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_delproduct
    op.add_column('batch_delproduct', sa.Column('inactivity_months', sa.Integer(), nullable=True))


def downgrade():

    # batch_delproduct
    op.drop_column('batch_delproduct', 'inactivity_months')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2022 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Commands relating to Postfix
"""

from __future__ import unicode_literals, absolute_import

import re
import sys
import subprocess
import logging

import six

from rattail.commands.core import Subcommand


log = logging.getLogger(__name__)


class PostfixSummary(Subcommand):
    """
    Generate (and maybe email) a Postfix log summary
    """
    name = 'postfix-summary'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--email', action='store_true', default=False,
                            help="Email the summary instead of displaying it.")

        parser.add_argument('--problems-only', action='store_true',
                            help="Only show summary if it contains problems.")

        parser.add_argument('--yesterday', action='store_true',
                            help="Show summary for previous day only.")

        parser.add_argument('file', nargs='*', default=['/var/log/mail.log'],
                            help="Path(s) to mail log file(s)")

    def run(self, args):

        # TODO: any way to get around using 'sudo' here?
        cmd = ['sudo', 'pflogsumm', '--problems-first']
        if args.yesterday:
            cmd.extend(['-d', 'yesterday'])
        cmd.extend(args.file)

        # run cmd
        output = subprocess.check_output(cmd).decode('utf_8')

        # parse problem counts
        problems = {}
        pattern = re.compile(r'^ +(\d)   (deferred|bounced|rejected|reject warnings|held|discarded)\b')
        for line in output.split('\n'):
            match = pattern.match(line)
            if match:
                problem = match.group(2)
                if problem in problems:
                    log.warning("extra line matches '%s' pattern: %s", problem, line)
                problems[problem] = int(match.group(1))

        log.info("postfix summary contained %s problems",
                 sum(six.itervalues(problems)))

        # bail if caller only wants to see problems and we have none
        if args.problems_only:
            if not any(six.itervalues(problems)):
                return

        # just display the summary unless we're to email it
        if args.email:
            self.app.send_email('postfix_summary', {
                'output': output,
                'problems': problems,
            })
        else:
            self.stdout.write(output)



# -*- coding: utf-8 -*-
"""add VendorSampleFile

Revision ID: e206b9457091
Revises: 31b0b868d6f0
Create Date: 2023-02-22 18:55:50.567493

"""

# revision identifiers, used by Alembic.
revision = 'e206b9457091'
down_revision = '31b0b868d6f0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor_sample_file
    op.create_table('vendor_sample_file',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('vendor_uuid', sa.String(length=32), nullable=False),
                    sa.Column('file_type', sa.String(length=100), nullable=False),
                    sa.Column('filename', sa.String(length=100), nullable=False),
                    sa.Column('bytes', sa.LargeBinary(), nullable=False),
                    sa.Column('effective_date', sa.Date(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'], name='vendor_sample_file_fk_user'),
                    sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'], name='vendor_sample_file_fk_vendor'),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('vendor_sample_file_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('vendor_uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('file_type', sa.String(length=100), autoincrement=False, nullable=False),
                    sa.Column('filename', sa.String(length=100), autoincrement=False, nullable=False),
                    sa.Column('effective_date', sa.Date(), autoincrement=False, nullable=True),
                    sa.Column('notes', sa.Text(), autoincrement=False, nullable=True),
                    sa.Column('created_by_uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_vendor_sample_file_version_end_transaction_id'), 'vendor_sample_file_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_vendor_sample_file_version_operation_type'), 'vendor_sample_file_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_vendor_sample_file_version_transaction_id'), 'vendor_sample_file_version', ['transaction_id'], unique=False)


def downgrade():

    # vendor_sample_file
    op.drop_index(op.f('ix_vendor_sample_file_version_transaction_id'), table_name='vendor_sample_file_version')
    op.drop_index(op.f('ix_vendor_sample_file_version_operation_type'), table_name='vendor_sample_file_version')
    op.drop_index(op.f('ix_vendor_sample_file_version_end_transaction_id'), table_name='vendor_sample_file_version')
    op.drop_table('vendor_sample_file_version')
    op.drop_table('vendor_sample_file')
